# © Kohulan Rajan - 2020
import tensorflow as tf
import os
from datetime import datetime
import pickle

print(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), "Process started", flush=True)

# Initial Setup for GPU
os.environ["CUDA_VISIBLE_DEVICES"] = "2"

gpus = tf.config.experimental.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)


def main():

    Smiles_Path = "Path_to_file"
    # Images_Path = "Train_Images/"

    img_name_vector, cap_vector, tokenizer, max_length, img_name_val = data_loader(
        Smiles_Path
    )

    print("Tokens: ", tokenizer, flush=True)
    print("Max length of attention weights: ", max_length, flush=True)
    print("Total train files:", len(img_name_vector), flush=True)

    with open("tokenizer_TPU_Stereo.pkl", "wb") as file:
        pickle.dump(tokenizer, file)

    with open("max_length_TPU_Stereo.pkl", "wb") as file:
        pickle.dump(max_length, file)

    start = 0
    end = 26800
    start_x = 0

    for i in range(int(len(img_name_vector) / end)):
        imgs_path = img_name_vector[start : (start + end)]
        caps_path = cap_vector[start : (start + end)]

        with open(
            "Images_Path_" + str(start_x) + "_" + str(start_x) + ".pkl", "wb"
        ) as file:
            pickle.dump(imgs_path, file)

        with open(
            "Capts_Path_" + str(start_x) + "_" + str(start_x) + ".pkl", "wb"
        ) as file:
            pickle.dump(caps_path, file)

        print("Total Train_Images: ", len(imgs_path), flush=True)
        print("Total SELFIES_Images: ", len(caps_path), flush=True)

        start = start + len(imgs_path)
        start_x = start_x + 1

    print(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), "Process completed", flush=True)


def data_loader(Smiles_Path):

    # read the Captions file
    with open(Smiles_Path, "r") as txt_file:
        smiles = txt_file.read()

    # storing the captions and the image name in vectors
    all_smiles = []
    all_img_name = []

    for line in smiles.split("\n"):
        # Split the ID and SMILES to seperate tokens
        tokens = line.split(",")

        image_id = str(tokens[0]) + ".png"
        # Add start and end annotations to SMILES string
        try:
            SMILES_ = (
                "<start> " + str(tokens[1].replace("][", "] [").rstrip()) + " <end>"
            )
        except IndexError as e:
            print(e, flush=True)

        # PATH of training images
        full_image_path = "Images_Path_/" + image_id

        all_img_name.append(full_image_path)
        all_smiles.append(SMILES_)

    train_captions, img_name_vector = (all_smiles, all_img_name)

    print("Total number of selected SMILES Strings: ", len(train_captions), flush=True)

    # Defining the maximum number of tokens to generate
    def calc_max_length(tensor):
        return max(len(t) for t in tensor)

    # choosing the top 500 words from the vocabulary
    top_k = 500
    tokenizer = tf.keras.preprocessing.text.Tokenizer(
        num_words=top_k, oov_token="<unk>", lower=False, filters=""
    )
    tokenizer.fit_on_texts(train_captions)
    train_seqs = tokenizer.texts_to_sequences(train_captions)

    tokenizer.word_index["<pad>"] = 0
    tokenizer.index_word[0] = "<pad>"

    # padding each vector to the max_length of the captions, if the max_length parameter is not provided, pad_sequences calculates that automatically
    cap_vector = tf.keras.preprocessing.sequence.pad_sequences(
        train_seqs, padding="post"
    )

    # calculating the max_length used to store the attention weights
    max_length = calc_max_length(train_seqs)

    img_name_train, img_name_val, cap_train, cap_val = (
        img_name_vector[0:1340000],
        img_name_vector[1340000:1440000],
        cap_vector[0:1340000],
        cap_vector[1340000:1440000],
    )

    print(
        str(len(img_name_train)),
        str(len(img_name_val)),
        str(len(cap_train)),
        str(len(cap_val)),
    )

    return img_name_train, cap_train, tokenizer, max_length, img_name_val


if __name__ == "__main__":
    main()



# -*- coding: UTF-8 -*-
# © Kohulan Rajan - 2020
import tensorflow as tf
import os
import pickle
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("file", nargs="+")

args = parser.parse_args()
print(args)
# Initial Setup for GPU
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

gpus = tf.config.experimental.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)

for file_in in args.file:

    def main():
        train_captions = pickle.load(
            open("Capts_Path_" + file_in + "_" + file_in + ".pkl", "rb")
        )
        img_name_vector = pickle.load(
            open("Images_Path_" + file_in + "_" + file_in + ".pkl", "rb")
        )

        print("Total number of selected SMILES Strings: ", len(train_captions), "\n")
        num_shards = int(len(train_captions) / 128)  # corresponds to total train files

        file_index = num_shards * int(file_in)

        get_train_tfrecord(num_shards, train_captions, img_name_vector, file_index)

    def _bytes_feature(value):
        """Returns a bytes_list from a string / byte."""
        if isinstance(value, type(tf.constant(0))):
            value = (
                value.numpy()
            )  # BytesList won't unpack a string from an EagerTensor.
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

    def get_train_tfrecord(num_shards, train_captions, img_name_vector, file_index):

        print("Total number of TFrecords: ", num_shards, flush=True)

        for i in range(num_shards):
            subsets_num = int(len(train_captions) / num_shards)
            sub_split_img_id = img_name_vector[i * subsets_num : (i + 1) * subsets_num]
            sub_split_cap_train = train_captions[
                i * subsets_num : (i + 1) * subsets_num
            ]

            tfrecord_name = "check/" + "train-%02d.tfrecord" % file_index
            writer = tf.io.TFRecordWriter(tfrecord_name)
            counter = 0
            for j in range(len(sub_split_img_id)):

                # print(decoded_image.shape)
                caption_ = sub_split_cap_train[j]
                # image_id_ = sub_split_img_id[counter]
                counter = counter + 1
                feature = {
                    # 'image_id': _bytes_feature(image_id_.encode('utf8')),
                    "image_raw": _bytes_feature(tf.io.read_file(sub_split_img_id[j])),
                    "caption": _bytes_feature(caption_.tostring()),
                }
                example = tf.train.Example(features=tf.train.Features(feature=feature))

                serialized = example.SerializeToString()
                writer.write(serialized)
            print("%s write to tfrecord success!" % tfrecord_name, flush=True)
            file_index = file_index + 1

    if __name__ == "__main__":
        main()



import re

pattern = "R([0-9]*)|X([0-9]*)|Y([0-9]*)|Z([0-9]*)"
add_space_re = "^(\W+)|(\W+)$"


def split_and_modify_atoms(SMILES):
    splitted_SMILES = list(SMILES)
    modified_SMILES = re.sub(r"\s+(?=[a-z])", "", " ".join(map(str, splitted_SMILES)))
    return modified_SMILES


def replacer(match):
    part = match.group(0)
    text = (
        part.replace("1", "!")
        .replace("2", "$")
        .replace("3", "^")
        .replace("4", "<")
        .replace("5", ">")
        .replace("6", "?")
        .replace("7", "£")
        .replace("8", "¢")
        .replace("9", "€")
        .replace("0", "§")
    )
    return text


def add_space(match_):
    if match_.group(1) is not None:
        return "{} ".format(match_.group(1))
    else:
        return " {}".format(match_.group(2))


def encoder(SMILES):
    replaced_SMILES = re.sub(pattern, replacer, SMILES)
    splitted_SMILES = split_and_modify_atoms(replaced_SMILES)
    modified_SMILES = " ".join(
        [re.sub(add_space_re, add_space, word) for word in splitted_SMILES.split()]
    )
    return modified_SMILES


def decoder(predictions):
    modified = (
        predictions.replace("!", "1")
        .replace("$", "2")
        .replace("^", "3")
        .replace("<", "4")
        .replace(">", "5")
        .replace("?", "6")
        .replace("£", "7")
        .replace("¢", "8")
        .replace("€", "9")
        .replace("§", "0")
    )
    return modified



import os
import sys
import logging
import pickle
import pystow
import tensorflow as tf
import DECIMER.config as config
import DECIMER.utils as utils

# Silence tensorflow model loading warnings.
logging.getLogger("absl").setLevel("ERROR")

# Silence tensorflow errors. optional not recommened if your model is not working properly.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

# Set the absolute path
HERE = os.path.dirname(os.path.abspath(__file__))

# Set model to run on default GPU and allow memory to grow as much as needed.
# This allows us to run multiple instances of inference in the same GPU.
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
gpus = tf.config.experimental.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)

# Set path
default_path = pystow.join("DECIMER-V2")

# model download location
model_url = "https://zenodo.org/record/7180845/files/models.zip"
model_path = str(default_path) + "/DECIMER_model/"

# download models to a default location
if not os.path.exists(model_path):
    config.download_trained_weights(model_url, default_path)


# Load important pickle files which consists the tokenizers and the maxlength setting

tokenizer = pickle.load(
    open(
        default_path.as_posix() + "/DECIMER_model/assets/tokenizer_SMILES.pkl",
        "rb",
    )
)


def main():
    """
    This function take the path of the image as user input
    and returns the predicted SMILES as output in CLI.

    Agrs:
        str: image_path

    Returns:
        str: predicted SMILES

    """
    if len(sys.argv) != 2:
        print("Usage: {} $image_path".format(sys.argv[0]))
    else:
        SMILES = predict_SMILES(sys.argv[1])
        print(SMILES)


def detokenize_output(predicted_array: int) -> str:
    """
    This function takes the predited tokens from the DECIMER model
    and returns the decoded SMILES string.

    Args:
        predicted_array (int): Predicted tokens from DECIMER

    Returns:
        (str): SMILES representation of the molecule
    """
    outputs = [tokenizer.index_word[i] for i in predicted_array[0].numpy()]
    prediction = (
        "".join([str(elem) for elem in outputs])
        .replace("<start>", "")
        .replace("<end>", "")
    )

    return prediction


# Load DECIMER model_packed
DECIMER_V2 = tf.saved_model.load(default_path.as_posix() + "/DECIMER_model/")


def predict_SMILES(image_path: str) -> str:
    """
    This function takes an image path (str) and returns the SMILES
    representation of the depicted molecule (str).

    Args:
        image_path (str): Path of chemical structure depiction image

    Returns:
        (str): SMILES representation of the molecule in the input image
    """
    chemical_structure = config.decode_image(image_path)
    predicted_tokens = DECIMER_V2(chemical_structure)
    predicted_SMILES = utils.decoder(detokenize_output(predicted_tokens))

    return predicted_SMILES


if __name__ == "__main__":
    main()



import tensorflow as tf
import numpy as np

# Set target dtype
TARGET_DTYPE = tf.float32


def get_angles(pos, i, d_model):
    angle_rates = tf.constant(1, TARGET_DTYPE) / tf.math.pow(
        tf.constant(10000, TARGET_DTYPE),
        (tf.constant(2, dtype=TARGET_DTYPE) * tf.cast((i // 2), TARGET_DTYPE)) / d_model,
    )
    return pos * angle_rates


def do_interleave(arr_a, arr_b):
    a_arr_tf_column = tf.range(arr_a.shape[1]) * 2  # [0 2 4 ...]
    b_arr_tf_column = tf.range(arr_b.shape[1]) * 2 + 1  # [1 3 5 ...]
    column_indices = tf.argsort(tf.concat([a_arr_tf_column, b_arr_tf_column], axis=-1))
    column, row = tf.meshgrid(column_indices, tf.range(arr_a.shape[0]))
    combine_indices = tf.stack([row, column], axis=-1)
    combine_value = tf.concat([arr_a, arr_b], axis=1)
    return tf.gather_nd(combine_value, combine_indices)


def positional_encoding_1d(position, d_model):
    angle_rads = get_angles(
        tf.cast(tf.range(position)[:, tf.newaxis], TARGET_DTYPE),
        tf.cast(tf.range(d_model)[tf.newaxis, :], TARGET_DTYPE),
        d_model,
    )

    # apply sin to even indices in the array; 2i
    sin_angle_rads = tf.math.sin(angle_rads[:, ::2])
    cos_angle_rads = tf.math.cos(angle_rads[:, 1::2])
    angle_rads = do_interleave(sin_angle_rads, cos_angle_rads)
    pos_encoding = angle_rads[tf.newaxis, ...]
    return pos_encoding


def np_positional_encoding_2d(row, col, d_model):
    assert d_model % 2 == 0
    row_pos = np.repeat(np.arange(row), col)[:, np.newaxis]
    col_pos = np.repeat(np.expand_dims(np.arange(col), 0), row, axis=0).reshape(-1, 1)

    angle_rads_row = get_angles(
        row_pos, np.arange(d_model // 2)[np.newaxis, :], d_model // 2
    ).numpy()
    angle_rads_col = get_angles(
        col_pos, np.arange(d_model // 2)[np.newaxis, :], d_model // 2
    ).numpy()

    angle_rads_row[:, 0::2] = np.sin(angle_rads_row[:, 0::2])
    angle_rads_row[:, 1::2] = np.cos(angle_rads_row[:, 1::2])
    angle_rads_col[:, 0::2] = np.sin(angle_rads_col[:, 0::2])
    angle_rads_col[:, 1::2] = np.cos(angle_rads_col[:, 1::2])
    pos_encoding = np.concatenate([angle_rads_row, angle_rads_col], axis=1)[
        np.newaxis, ...
    ]
    return tf.cast(pos_encoding, dtype=TARGET_DTYPE)


def positional_encoding_2d(row, col, d_model):
    row_pos = tf.repeat(tf.range(row), col)[:, tf.newaxis]
    col_pos = tf.reshape(
        tf.repeat(tf.expand_dims(tf.range(col), 0), row, axis=0), (-1, 1)
    )

    angle_rads_row = get_angles(
        tf.cast(row_pos, tf.float32),
        tf.range(d_model // 2)[tf.newaxis, :],
        d_model // 2,
    )
    angle_rads_col = get_angles(
        tf.cast(col_pos, tf.float32),
        tf.range(d_model // 2)[tf.newaxis, :],
        d_model // 2,
    )

    sin_angle_rads_row = tf.math.sin(angle_rads_row[:, ::2])
    cos_angle_rads_row = tf.math.cos(angle_rads_row[:, 1::2])
    angle_rads_row = do_interleave(sin_angle_rads_row, cos_angle_rads_row)

    sin_angle_rads_col = tf.math.sin(angle_rads_col[:, ::2])
    cos_angle_rads_col = tf.math.cos(angle_rads_col[:, 1::2])
    angle_rads_col = do_interleave(sin_angle_rads_col, cos_angle_rads_col)

    pos_encoding = tf.concat([angle_rads_row, angle_rads_col], axis=1)[tf.newaxis, ...]
    return pos_encoding


def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), TARGET_DTYPE)

    # add extra dimensions to add the padding to the attention logits.
    #    - (batch_size, 1, 1, seq_len)
    return seq[:, tf.newaxis, tf.newaxis, :]


def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    # (seq_len, seq_len)
    return tf.cast(mask, TARGET_DTYPE)


def create_mask(inp, tar):

    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)
    return tf.cast(combined_mask, TARGET_DTYPE)


def scaled_dot_product_attention(q, k, v, mask):
    matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)

    # scale matmul_qk
    dk = tf.cast(tf.shape(k)[-1], TARGET_DTYPE)

    # Calculate scaled attention logits
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

    # add the mask to the scaled tensor.
    if mask is not None:
        scaled_attention_logits += mask * -1e9

    # softmax is normalized on the last axis (seq_len_k)
    # so that the scores add up to 1.
    #   - shape --> (..., seq_len_q, seq_len_k)
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)

    #   - shape --> (..., seq_len_q, depth_v)
    output = tf.matmul(attention_weights, v)

    return output, attention_weights


class MultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads):
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)

        self.dense = tf.keras.layers.Dense(d_model)

    def split_heads(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        # (batch_size, seq_len, d_model)
        q = self.wq(q)
        # (batch_size, seq_len, d_model)
        k = self.wk(k)
        # (batch_size, seq_len, d_model)
        v = self.wv(v)

        # (batch_size, num_heads, seq_len_q, depth)
        q = self.split_heads(q, batch_size)
        # (batch_size, num_heads, seq_len_k, depth)
        k = self.split_heads(k, batch_size)
        # (batch_size, num_heads, seq_len_v, depth)
        v = self.split_heads(v, batch_size)

        # scaled_attention.shape – (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape – (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = scaled_dot_product_attention(
            q, k, v, mask
        )

        # (batch_size, seq_len_q, num_heads, depth)
        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])

        # (batch_size, seq_len_q, d_model)
        concat_attention = tf.reshape(scaled_attention, (batch_size, -1, self.d_model))

        # (batch_size, seq_len_q, d_model)
        output = self.dense(concat_attention)

        return output, attention_weights


def point_wise_feed_forward_network(d_model, dff):
    return tf.keras.Sequential(
        [
            # INNER LAYER
            #   – (batch_size, seq_len, dff)
            tf.keras.layers.Dense(dff, activation="relu"),
            # OUTPUT
            #   – (batch_size, seq_len, d_model)
            tf.keras.layers.Dense(d_model),
        ]
    )


class TransformerEncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, dropout_rate=0.1):

        super(TransformerEncoderLayer, self).__init__()

        self.mha = tf.keras.layers.MultiHeadAttention(
            num_heads,
            key_dim=d_model,
        )
        self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout2 = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x, training, mask=None):

        # returns (batch_size, input_seq_len, d_model)
        attn_output, _ = self.mha(x, x, x, mask, return_attention_scores=True)

        # Potentially unncessary by passing dropout1 to tf.keras.layers.MultiHeadAttention (if using tf MHA)
        attn_output = self.dropout1(attn_output, training=training)

        # Residual connection followed by layer normalization
        #   – returns (batch_size, input_seq_len, d_model)
        out1 = self.layernorm1(x + attn_output, training=training)

        # Point-wise Feed Forward Step
        #   – returns (batch_size, input_seq_len, d_model)
        ffn_output = self.ffn(out1, training=training)
        ffn_output = self.dropout2(ffn_output, training=training)

        # Residual connection followed by layer normalization
        #   – returns (batch_size, input_seq_len, d_model)
        out2 = self.layernorm2(out1 + ffn_output, training=training)

        return out2


class TransformerDecoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, dropout_rate=0.1):

        super(TransformerDecoderLayer, self).__init__()

        # WE COULD USE A CUSTOM DEFINED MHA MODEL BUT WE WILL USE TFA INSTEAD
        self.mha1 = MultiHeadAttention(d_model, num_heads)
        self.mha2 = MultiHeadAttention(d_model, num_heads)
        #
        # # Multi Head Attention Layers
        # self.mha1 = tf.keras.layers.MultiHeadAttention(num_heads, key_dim=d_model,)
        # self.mha2 = tf.keras.layers.MultiHeadAttention(num_heads, key_dim=d_model,)

        # Feed Forward NN
        self.ffn = point_wise_feed_forward_network(d_model, dff)

        # Layer Normalization Layers
        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm3 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        # Dropout Layers
        self.dropout1 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout2 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout3 = tf.keras.layers.Dropout(dropout_rate)

    # enc_output.shape == (batch_size, input_seq_len, d_model)
    def call(self, x, enc_output, training, look_ahead_mask=None, padding_mask=None):

        attn1, attn_weights_block1 = self.mha1(x, x, x, look_ahead_mask)
        attn1 = self.dropout1(attn1, training=training)

        # Residual connection followed by layer normalization
        #   – (batch_size, target_seq_len, d_model)
        out1 = self.layernorm1(attn1 + x, training=training)

        # Merging connection between encoder and decoder (MHA)
        #   – (batch_size, target_seq_len, d_model)
        attn2, attn_weights_block2 = self.mha2(
            enc_output, enc_output, out1, padding_mask
        )
        attn2 = self.dropout2(attn2, training=training)

        # Residual connection followed by layer normalization
        #   – (batch_size, target_seq_len, d_model)
        out2 = self.layernorm2(attn2 + out1, training=training)

        # (batch_size, target_seq_len, d_model)
        ffn_output = self.ffn(out2, training=training)
        ffn_output = self.dropout3(ffn_output, training=training)

        # Residual connection followed by layer normalization
        #   – (batch_size, target_seq_len, d_model)
        out3 = self.layernorm3(ffn_output + out2, training=training)

        return out3, attn_weights_block1, attn_weights_block2


class TransformerEncoder(tf.keras.layers.Layer):
    def __init__(
        self,
        num_layers,
        d_model,
        num_heads,
        dff,
        maximum_position_encoding,
        dropout_rate=0.1,
    ):
        super(TransformerEncoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers
        self.pos_encoding = positional_encoding_1d(
            maximum_position_encoding, self.d_model
        )
        self.enc_layers = [
            TransformerEncoderLayer(d_model, num_heads, dff, dropout_rate)
            for _ in range(num_layers)
        ]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x, training, mask=None):
        # adding embedding and position encoding.
        #   – (batch_size, input_seq_len, d_model)
        x += self.pos_encoding
        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, mask)

        #   – (batch_size, input_seq_len, d_model)
        return x


class TransformerDecoder(tf.keras.layers.Layer):
    def __init__(
        self,
        num_layers,
        d_model,
        num_heads,
        dff,
        target_vocab_size,
        maximum_position_encoding,
        dropout_rate=0.1,
    ):
        super(TransformerDecoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoding = positional_encoding_1d(maximum_position_encoding, d_model)

        self.dec_layers = [
            TransformerDecoderLayer(d_model, num_heads, dff, dropout_rate)
            for _ in range(num_layers)
        ]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x, enc_output, training, look_ahead_mask=None, padding_mask=None):

        seq_len = tf.shape(x)[1]
        attention_weights = {}

        # adding embedding and position encoding.
        #   – (batch_size, target_seq_len, d_model)
        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.d_model, TARGET_DTYPE))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](
                x, enc_output, training, look_ahead_mask, padding_mask
            )
            attention_weights["decoder_layer{}_block1".format(i + 1)] = block1
            attention_weights["decoder_layer{}_block2".format(i + 1)] = block2

        # x.shape == (batch_size, target_seq_len, d_model)
        return x, attention_weights


class Transformer(tf.keras.Model):
    def __init__(
        self,
        num_layers,
        d_model,
        num_heads,
        dff,
        target_vocab_size,
        pe_input,
        pe_target,
        dropout_rate=0.1,
    ):

        super(Transformer, self).__init__()

        self.t_encoder = TransformerEncoder(
            num_layers, d_model, num_heads, dff, pe_input, dropout_rate
        )
        self.t_decoder = TransformerDecoder(
            num_layers,
            d_model,
            num_heads,
            dff,
            target_vocab_size,
            pe_target,
            dropout_rate,
        )
        self.t_final_layer = tf.keras.layers.Dense(target_vocab_size)

    def call(
        self,
        t_inp,
        t_tar,
        training,
        enc_padding_mask=None,
        look_ahead_mask=None,
        dec_padding_mask=None,
    ):
        # (batch_size, inp_seq_len, d_model)
        enc_output = self.t_encoder(t_inp, training, enc_padding_mask)

        # dec_output.shape == (batch_size, tar_seq_len, d_model)
        dec_output, attention_weights = self.t_decoder(
            t_tar, enc_output, training, look_ahead_mask, dec_padding_mask
        )

        # (batch_size, tar_seq_len, target_vocab_size)
        final_output = self.t_final_layer(dec_output)

        return final_output, attention_weights



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Hparams for model architecture and trainer."""
import ast
import collections
import copy
from typing import Any, Dict, Text
import tensorflow as tf
import yaml


def eval_str_fn(val):
    if "|" in val:
        return [eval_str_fn(v) for v in val.split("|")]
    if val in {"true", "false"}:
        return val == "true"
    try:
        return ast.literal_eval(val)
    except (ValueError, SyntaxError):
        return val


# pylint: disable=protected-access
class Config(dict):
    """A config utility class."""

    def __init__(self, *args, **kwargs):
        super().__init__()
        input_config_dict = dict(*args, **kwargs)
        self.update(input_config_dict)

    def __len__(self):
        return len(self.__dict__)

    def __setattr__(self, k, v):
        if isinstance(v, dict) and not isinstance(v, Config):
            self.__dict__[k] = Config(v)
        else:
            self.__dict__[k] = copy.deepcopy(v)

    def __getattr__(self, k):
        return self.__dict__[k]

    def __setitem__(self, k, v):
        self.__setattr__(k, v)

    def __getitem__(self, k):
        return self.__dict__[k]

    def __iter__(self):
        for key in self.__dict__:
            yield key

    def items(self):
        for key, value in self.__dict__.items():
            yield key, value

    def __repr__(self):
        return repr(self.as_dict())

    def __getstate__(self):
        return self.__dict__

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            result[k] = v
        return result

    def __str__(self):
        try:
            return yaml.dump(self.as_dict(), indent=4)
        except TypeError:
            return str(self.as_dict())

    def _update(self, config_dict, allow_new_keys=True):
        """Recursively update internal members."""
        if not config_dict:
            return

        for k, v in config_dict.items():
            if k not in self.__dict__:
                if allow_new_keys:
                    self.__setattr__(k, v)
                else:
                    raise KeyError("Key `{}` does not exist for overriding. ".format(k))
            else:
                if isinstance(self.__dict__[k], Config) and isinstance(v, dict):
                    self.__dict__[k]._update(v, allow_new_keys)
                elif isinstance(self.__dict__[k], Config) and isinstance(v, Config):
                    self.__dict__[k]._update(v.as_dict(), allow_new_keys)
                else:
                    self.__setattr__(k, v)

    def get(self, k, default_value=None):
        return self.__dict__.get(k, default_value)

    def update(self, config_dict):
        """Update members while allowing new keys."""
        self._update(config_dict, allow_new_keys=True)

    def keys(self):
        return self.__dict__.keys()

    def override(self, config_dict_or_str, allow_new_keys=False):
        """Update members while disallowing new keys."""
        if not config_dict_or_str:
            return
        if isinstance(config_dict_or_str, str):
            if "=" in config_dict_or_str:
                config_dict = self.parse_from_str(config_dict_or_str)
            elif config_dict_or_str.endswith(".yaml"):
                config_dict = self.parse_from_yaml(config_dict_or_str)
            else:
                raise ValueError(
                    'Invalid string {}, must end with .yaml or contains "=".'.format(
                        config_dict_or_str
                    )
                )
        elif isinstance(config_dict_or_str, dict):
            config_dict = config_dict_or_str
        else:
            raise ValueError("Unknown value type: {}".format(config_dict_or_str))

        self._update(config_dict, allow_new_keys)

    def parse_from_yaml(self, yaml_file_path: Text) -> Dict[Any, Any]:
        """Parses a yaml file and returns a dictionary."""
        with tf.io.gfile.GFile(yaml_file_path, "r") as f:
            config_dict = yaml.load(f, Loader=yaml.FullLoader)
            return config_dict

    def save_to_yaml(self, yaml_file_path):
        """Write a dictionary into a yaml file."""
        with tf.io.gfile.GFile(yaml_file_path, "w") as f:
            yaml.dump(self.as_dict(), f, default_flow_style=False)

    def parse_from_str(self, config_str: Text) -> Dict[Any, Any]:
        """Parse a string like 'x.y=1,x.z=2' to nested dict {x: {y: 1, z: 2}}."""
        if not config_str:
            return {}
        config_dict = {}
        try:
            for kv_pair in config_str.split(","):
                if not kv_pair:  # skip empty string
                    continue
                key_str, value_str = kv_pair.split("=")
                key_str = key_str.strip()

                def add_kv_recursive(k, v):
                    """Recursively parse x.y.z=tt to {x: {y: {z: tt}}}."""
                    if "." not in k:
                        return {k: eval_str_fn(v)}
                    pos = k.index(".")
                    return {k[:pos]: add_kv_recursive(k[pos + 1 :], v)}

                def merge_dict_recursive(target, src):
                    """Recursively merge two nested dictionary."""
                    for k in src.keys():
                        if (
                            k in target
                            and isinstance(target[k], dict)
                            and isinstance(src[k], collections.abc.Mapping)
                        ):
                            merge_dict_recursive(target[k], src[k])
                        else:
                            target[k] = src[k]

                merge_dict_recursive(config_dict, add_kv_recursive(key_str, value_str))
            return config_dict
        except ValueError:
            raise ValueError("Invalid config_str: {}".format(config_str))

    def as_dict(self):
        """Returns a dict representation."""
        config_dict = {}
        for k, v in self.__dict__.items():
            if isinstance(v, Config):
                config_dict[k] = v.as_dict()
            elif isinstance(v, (list, tuple)):
                config_dict[k] = [
                    i.as_dict() if isinstance(i, Config) else copy.deepcopy(i)
                    for i in v
                ]
            else:
                config_dict[k] = copy.deepcopy(v)
        return config_dict
        # pylint: enable=protected-access


registry_map = {}


def register(cls, prefix="effnet:"):
    """Register a function, mainly for config here."""
    registry_map[prefix + cls.__name__.lower()] = cls
    return cls


def lookup(name, prefix="effnet:") -> Any:
    name = prefix + name.lower()
    if name not in registry_map:
        raise ValueError(f"{name} not registered: {registry_map.keys()}")
    return registry_map[name]


base_config = Config(
    # model related params.
    model=dict(
        model_name="efficientnet-b0",
        data_format="channels_last",
        feature_size=1280,
        bn_type=None,  # 'tpu_bn',
        bn_momentum=0.9,
        bn_epsilon=1e-3,
        gn_groups=8,
        depth_divisor=8,
        min_depth=8,
        act_fn="silu",
        survival_prob=0.8,
        local_pooling=False,
        headbias=None,
        conv_dropout=None,
        dropout_rate=None,
        depth_coefficient=None,
        width_coefficient=None,
        blocks_args=None,
        num_classes=1000,  # must be the same as data.num_classes
    ),
    # train related params.
    train=dict(
        stages=0,
        epochs=350,
        min_steps=0,
        optimizer="rmsprop",
        lr_sched="exponential",
        lr_base=0.016,
        lr_decay_epoch=2.4,
        lr_decay_factor=0.97,
        lr_warmup_epoch=5,
        lr_min=0,
        ema_decay=0.9999,
        weight_decay=1e-5,
        weight_decay_inc=0.0,
        weight_decay_exclude=".*(bias|gamma|beta).*",
        label_smoothing=0.1,
        gclip=0,
        batch_size=4096,
        isize=None,
        split=None,  # dataset split, default to 'train'
        loss_type=None,  # loss type: sigmoid or softmax
        ft_init_ckpt=None,
        ft_init_ema=True,
        varsexp=None,  # trainable variables.
        sched=None,  # schedule
    ),
    eval=dict(
        batch_size=8,
        isize=None,  # image size
        split=None,  # dataset split, default to 'eval'
    ),
    # data related params.
    data=dict(
        ds_name="imagenet",
        augname="randaug",  # or 'autoaug'
        ra_num_layers=2,
        ram=15,
        mixup_alpha=0.0,
        cutmix_alpha=0.0,
        ibase=128,
        cache=True,
        resize=None,
        data_dir=None,
        multiclass=None,
        num_classes=1000,
        tfds_name=None,
        tfds_split=None,
        splits=dict(
            train=dict(num_images=None, files=None, tfds_split=None, slice=None),
            eval=dict(num_images=None, files=None, tfds_split=None, slice=None),
            minival=dict(num_images=None, files=None, tfds_split=None, slice=None),
            trainval=dict(num_images=None, files=None, tfds_split=None, slice=None),
        ),
    ),
    runtime=dict(
        iterations_per_loop=1000,  # larger value has better utilization.
        skip_host_call=False,
        mixed_precision=True,
        use_async_checkpointing=False,
        log_step_count_steps=64,
        keep_checkpoint_max=5,
        keep_checkpoint_every_n_hours=5,
    ),
)



import config
import Efficient_Net_encoder
import Transformer_decoder
import efficientnet.tfkeras as efn
import time
import re
import sys
import pickle
from datetime import datetime
import matplotlib.pyplot as plt
import tensorflow as tf
import matplotlib as mpl

mpl.use("Agg")

# Set TPUs

tpu = tf.distribute.cluster_resolver.TPUClusterResolver(tpu="node-name")
print("Running on TPU ", tpu.master())

tf.config.experimental_connect_to_cluster(tpu)
tf.tpu.experimental.initialize_tpu_system(tpu)
strategy = tf.distribute.TPUStrategy(tpu)

print("Number of devices: {}".format(strategy.num_replicas_in_sync), flush=True)
print("REPLICAS: ", strategy.num_replicas_in_sync, flush=True)
print(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), "Network Started", flush=True)

# Load the Data
total_data = 1000000  # datasize integer

tokenizer = pickle.load(open("tokenizer_Isomeric_SMILES.pkl", "rb"))
max_length = pickle.load(open("max_length_Isomeric_SMILES.pkl", "rb"))

PAD_TOKEN = tf.constant(tokenizer.word_index["<pad>"], dtype=tf.int32)

# Image parameters
IMG_EMB_DIM = (10, 10, 232)
IMG_EMB_DIM = (IMG_EMB_DIM[0] * IMG_EMB_DIM[1], IMG_EMB_DIM[2])
IMG_SHAPE = (299, 299, 3)
PE_INPUT = IMG_EMB_DIM[0]
IMG_SEQ_LEN, IMG_EMB_DEPTH = IMG_EMB_DIM
D_MODEL = IMG_EMB_DEPTH

# Set Training Epochs
EPOCHS = 40
REPLICA_BATCH_SIZE = 128
BATCH_SIZE = REPLICA_BATCH_SIZE * strategy.num_replicas_in_sync
print(BATCH_SIZE)
BUFFER_SIZE = 10000
target_vocab_size = max_length
TRAIN_STEPS = total_data // BATCH_SIZE
validation_steps = 15360 // BATCH_SIZE

# Parameters to train the network
N_LAYERS = 4
D_MODEL = 512
D_FF = 2048
N_HEADS = 8
DROPOUT_RATE = 0.1

# ENCODER_CONFIG
PREPROCESSING_FN = tf.keras.applications.efficientnet.preprocess_input
BB_FN = Efficient_Net_encoder.get_efficientnetv2_backbone

MAX_LEN = max_length
VOCAB_LEN = len(tokenizer.word_index)
PE_OUTPUT = MAX_LEN
TARGET_V_SIZE = len(tokenizer.word_index)
WARM_STEPS = 4000
TARGET_DTYPE = tf.float32

print("Total train steps: ", TRAIN_STEPS)
AUTO = tf.data.experimental.AUTOTUNE


def decode_image(image_data):
    """Preprocess the input image for Efficient-Net and
    returned the preprocessed image

    Args:
        image_data (int array): Decoded image in 2D array

    Returns:
        image (array): Preprocessed image in 2D array
    """
    try:
        img = tf.image.decode_png(image_data, channels=3)
    except InvalidArgumentError as e:
        print(e)
        pass
    img = tf.image.resize(img, (299, 299))
    img = efn.preprocess_input(img)
    # img = tf.expand_dims(img, 0)
    # print(img)
    return img


def read_tfrecord(example):
    """Read a tf record file and decodes the image and text data
    back into original form.

    Args:
        example (tf.record): single entry from tf record file

    Returns:
        img (float array): 2D float array
        caption: tokenized SMILES string
    """
    feature = {
        "image_raw": tf.io.FixedLenFeature([], tf.string),
        "caption": tf.io.FixedLenFeature([], tf.string),
    }

    # decode the TFRecord
    example = tf.io.parse_single_example(example, feature)
    img = decode_image(example["image_raw"])
    caption = tf.io.decode_raw(example["caption"], tf.int32)
    return img, caption


numbers = re.compile(r"(\d+)")


def numericalSort(value):
    """Sorts the filenames numerically

    Args:
        value (int): numerical value of the file name

    Returns:
        parts (int): sorted value
    """
    parts = numbers.split(value)
    parts[1::2] = map(int, parts[1::2])
    return parts


def get_dataset(batch_size=BATCH_SIZE, buffered_size=BUFFER_SIZE, path=""):
    """Creates a batch of data rom a given dataset

    Args:
        batch_size (int, optional): number of datapoints per batch. Defaults to BATCH_SIZE.
        buffered_size (int, optional): number of datapoints to be buffered. Defaults to BUFFER_SIZE.
        path (str, optional): path to the location of the dataset. Defaults to "".

    Returns:
        train_dataset (): single batch of datapoints
    """
    options = tf.data.Options()
    # filenames = tf.io.gfile.glob(path)
    filenames = sorted(tf.io.gfile.glob(path), key=numericalSort)
    print(len(filenames))
    fx = open("filenames", "w")
    for i in range(len(filenames)):
        fx.write(filenames[i] + "\n")
    fx.close()

    dataset_img = tf.data.TFRecordDataset(filenames, num_parallel_reads=AUTO)

    train_dataset = (
        dataset_img.with_options(options)
        .map(read_tfrecord, num_parallel_calls=AUTO)
        .repeat()
        .shuffle(buffered_size)
        .batch(batch_size, drop_remainder=True)
        .prefetch(buffer_size=AUTO)
    )
    return train_dataset


train_dataset = strategy.experimental_distribute_dataset(
    get_dataset(path="gs://tpu-test-koh/DECIMER_V2/RanDepict/*.tfrecord")
)
# validation_dataset = strategy.experimental_distribute_dataset(get_dataset(path="gs://tpu-test-koh/DECIMER_V2/RanDepict/Val_data/*.tfrecord"))


# validation_dataset = strategy.experimental_distribute_dataset(get_validation_dataset())

training_config = config.Config()
training_config.initialize_transformer_config(
    vocab_len=VOCAB_LEN,
    max_len=MAX_LEN,
    n_transformer_layers=N_LAYERS,
    transformer_d_dff=D_FF,
    transformer_n_heads=N_HEADS,
    image_embedding_dim=IMG_EMB_DIM,
)
training_config.initialize_encoder_config(
    image_embedding_dim=IMG_EMB_DIM,
    preprocessing_fn=PREPROCESSING_FN,
    backbone_fn=BB_FN,
    image_shape=IMG_SHAPE,
    do_permute=IMG_EMB_DIM[1] < IMG_EMB_DIM[0],
)
training_config.initialize_lr_config(
    warm_steps=WARM_STEPS,
    n_epochs=EPOCHS,
)

print(f"\n Encoder config:\n\t -> {training_config.encoder_config}\n")
print(f"\n Transformer config:\n\t -> {training_config.transformer_config}\n")
print(f"Learning rate config:\n\t -> {training_config.lr_config}\n")

print("Training preparation\n")


def prepare_for_training(lr_config, encoder_config, transformer_config, verbose=0):
    """Preparte the model for training. initiate the learning rate, loss object, metrics
    and optimizer

    Args:
        lr_config (int): values for learning rate configuration
        encoder_config (_type_): encoder configuration values
        transformer_config (_type_): transformer configuration values
        verbose (int, optional): _description_. Defaults to 0.

    Returns:
        loss_function, optimizer, model and the metrics
    """

    with strategy.scope():

        loss_object = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.NONE
        )

        def loss_fn(real, pred):
            # Convert to uint8
            mask = tf.math.logical_not(tf.math.equal(real, 0))
            loss_ = loss_object(real, pred)
            loss_ *= tf.cast(mask, dtype=loss_.dtype)
            loss_ = tf.nn.compute_average_loss(
                loss_, global_batch_size=REPLICA_BATCH_SIZE
            )
            return loss_

        train_loss = tf.keras.metrics.Mean(name="train_loss", dtype=tf.float32)
        train_accuracy = tf.keras.metrics.SparseCategoricalAccuracy(
            name="train_accuracy", dtype=tf.float32
        )
        """
        validation_loss = tf.keras.metrics.Mean(
            name="validation_loss", dtype=tf.float32
        )
        validation_accuracy = tf.keras.metrics.SparseCategoricalAccuracy(
            name="validation_accuracy", dtype=tf.float32
        )
        """
        # Declare the learning rate schedule (try this as actual lr schedule and list...)
        lr_scheduler = config.CustomSchedule(
            transformer_config["d_model"], lr_config["warm_steps"]
        )

        # Instiate an optimizer
        optimizer = tf.keras.optimizers.Adam(
            lr_scheduler, beta_1=0.9, beta_2=0.98, epsilon=1e-9
        )

        # Instantiate the encoder model
        encoder = Efficient_Net_encoder.Encoder(**encoder_config)
        initialization_batch = encoder(
            tf.ones(
                ((REPLICA_BATCH_SIZE,) + encoder_config["image_shape"]),
                dtype=TARGET_DTYPE,
            ),
            training=False,
        )

        # Instantiate the decoder model
        transformer = Transformer_decoder.Transformer(**transformer_config)
        transformer(
            initialization_batch,
            tf.random.uniform((REPLICA_BATCH_SIZE, 1)),
            training=False,
        )

    # Show the model architectures and plot the learning rate
    if verbose:
        print("\nEncoder model\n")
        print(encoder.summary())

        print("\nTransformer model\n")
        print(transformer.summary())

    return (
        loss_fn,
        optimizer,
        lr_scheduler,
        encoder,
        transformer,
        train_loss,
        train_accuracy,
    )


# Instantiate our required training components in the correct scope
(
    loss_fn,
    optimizer,
    lr_scheduler,
    encoder,
    transformer,
    train_loss,
    train_accuracy,
) = prepare_for_training(
    lr_config=training_config.lr_config,
    encoder_config=training_config.encoder_config,
    transformer_config=training_config.transformer_config,
    verbose=1,
)

print("\n Training preparation completed\n")

# Path to checkpoints
checkpoint_path = "gs://tpu-test-koh/DECIMER_V2/checkpoints_SMILES_TPU8/"
ckpt = tf.train.Checkpoint(
    encoder=encoder, transformer=transformer, optimizer=optimizer
)
ckpt_manager = tf.train.CheckpointManager(ckpt, checkpoint_path, max_to_keep=50)

start_epoch = 0
if ckpt_manager.latest_checkpoint:
    ckpt.restore(tf.train.latest_checkpoint(checkpoint_path))
    start_epoch = int(ckpt_manager.latest_checkpoint.split("-")[-1])


# Main training step fucntion
def train_step(image_batch, selfies_batch):
    """Main training step function

    Args:
        image_batch (float array): Input image batch
        selfies_batch (int array): Input selfies batch tokenized
    """

    selfies_batch_input = selfies_batch[:, :-1]
    selfies_batch_target = selfies_batch[:, 1:]
    combined_mask = Transformer_decoder.create_mask(
        selfies_batch_input, selfies_batch_target
    )

    with tf.GradientTape() as tape:
        image_embedding = encoder(image_batch, training=True)
        prediction_batch, _ = transformer(
            image_embedding,
            selfies_batch_input,
            training=True,
            look_ahead_mask=combined_mask,
        )

        # Update Loss Accumulator
        batch_loss = loss_fn(selfies_batch_target, prediction_batch) / (MAX_LEN - 1)
        train_accuracy.update_state(
            selfies_batch_target,
            prediction_batch,
            sample_weight=tf.where(
                tf.not_equal(selfies_batch_target, PAD_TOKEN), 1.0, 0.0
            ),
        )

    # total_loss = batch_loss / int(selfies_batch.shape[1])

    gradients = tape.gradient(
        batch_loss, encoder.trainable_variables + transformer.trainable_variables
    )
    gradients, _ = tf.clip_by_global_norm(gradients, 10.0)
    optimizer.apply_gradients(
        zip(gradients, encoder.trainable_variables + transformer.trainable_variables)
    )

    train_loss.update_state(batch_loss * strategy.num_replicas_in_sync)


@tf.function
def dist_train_step(image_batch, selfies_batch):
    strategy.run(train_step, args=(image_batch, selfies_batch))


"""
# Main validation step fucntion
def validation_step(image_batch, selfies_batch):

    selfies_batch_input = selfies_batch[:, :-1]
    selfies_batch_target = selfies_batch[:, 1:]
    combined_mask = create_mask(selfies_batch_input, selfies_batch_target)

    with tf.GradientTape() as tape:
        image_embedding = encoder(image_batch, training=True)
        prediction_batch, _ = transformer(
            image_embedding,
            selfies_batch_input,
            training=True,
            look_ahead_mask=combined_mask,
        )

        # Update Loss Accumulator
        batch_loss = loss_fn(selfies_batch_target, prediction_batch) / (MAX_LEN - 1)
        validation_accuracy.update_state(selfies_batch_target, prediction_batch,sample_weight=tf.where(tf.not_equal(selfies_batch_target, PAD_TOKEN), 1.0, 0.0))

    total_loss = batch_loss / int(selfies_batch.shape[1])

    gradients = tape.gradient(
        batch_loss, encoder.trainable_variables + transformer.trainable_variables
    )
    gradients, _ = tf.clip_by_global_norm(gradients, 10.0)
    optimizer.apply_gradients(
        zip(gradients, encoder.trainable_variables + transformer.trainable_variables)
    )

    train_loss.update_state(batch_loss * strategy.num_replicas_in_sync)


@tf.function
def dist_validation_step(image_batch, selfies_batch):
    strategy.run(
        validation_step, args=(image_batch, selfies_batch)
    )
"""
loss_plot = []
accuracy_plot = []
val_loss_plot = []
val_acc = []
f = open("Training_SMILES_TPU8__.txt", "w")
sys.stdout = f

# Training loop
for epoch in range(start_epoch, EPOCHS):
    start = time.time()
    batch = 0
    validation_batch = 0

    for x in train_dataset:
        img_tensor, target = x
        dist_train_step(img_tensor, target)
        batch += 1

        if batch % 100 == 0:
            print(
                "Epoch {} Batch {} Loss {:.4f} Accuracy {:.4f}".format(
                    epoch + 1, batch, train_loss.result(), train_accuracy.result()
                ),
                flush=True,
            )
        # storing the epoch end loss value to plot later

        if batch == TRAIN_STEPS:
            loss_plot.append(train_loss.result().numpy())
            accuracy_plot.append(train_accuracy.result().numpy())
            ckpt_manager.save()

            print(
                "Epoch {} Training_Loss {:.4f} Accuracy {:.4f}".format(
                    epoch + 1,
                    train_loss.result(),
                    train_accuracy.result(),
                ),
                flush=True,
            )
            print(
                datetime.now().strftime("%Y/%m/%d %H:%M:%S"),
                "Time taken for 1 epoch {} sec\n".format(time.time() - start),
                flush=True,
            )
            # transformer.save_weights('Epoch_'+str(epoch+1)+'_weights.h5')

            break
    """
    for y in validation_dataset:
        start_val = time.time()
        img_tensor, target =y
        validation_step(y)
        validation_batch +=1

        if validation_batch == validation_steps:

            print ('Validation_Loss {:.4f} Accuracy {:.4f}'.format(validation_loss.result(), validation_accuracy.result()), flush=True)
            print (datetime.now().strftime('%Y/%m/%d %H:%M:%S'),'Time taken for validation {} sec\n'.format(time.time() - start_val), flush=True)
            val_loss_plot.append(validation_loss.result().numpy())
            val_acc.append(validation_accuracy.result().numpy())
            break
    """
    train_loss.reset_states()
    train_accuracy.reset_states()
    # validation_loss.reset_states()
    # validation_accuracy.reset_states()

    # epoch = (epoch+1)
    # batch = 0

plt.plot(loss_plot, "-o", label="Training loss")
# plt.plot(val_loss_plot , '-o', label= "Validation loss")
plt.title("Training and Validation Loss")
plt.xlabel("Epochs")
plt.ylabel("Loss")
plt.legend(ncol=2, loc="upper right")
# plt.show()
plt.gcf().set_size_inches(20, 20)
plt.savefig("Lossplot_SMILESTPU8.jpg")
plt.close()

plt.plot(accuracy_plot, "-o", label="Training accuracy")
# plt.plot(val_acc , '-o', label= "Validation accuracy")
plt.title("Training and Validation accuracy")
plt.xlabel("Epochs")
plt.ylabel("Loss")
plt.legend(ncol=2, loc="lower right")
# plt.show()
plt.gcf().set_size_inches(20, 20)
plt.savefig("accuracyplot_SMILESTPU8.jpg")
plt.close()

print(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), "Network Completed", flush=True)
f.close()



import os
import tensorflow as tf

import pickle
import Transformer_decoder
import Efficient_Net_encoder
import config

print(tf.__version__)

# Set GPU
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
gpus = tf.config.experimental.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)

# load assets
tokenizer = pickle.load(open("tokenizer_TPU_Stereo.pkl", "rb"))
max_length = pickle.load(open("max_length_TPU_Stereo.pkl", "rb"))

# Image parameters
IMG_EMB_DIM = (10, 10, 232)
IMG_EMB_DIM = (IMG_EMB_DIM[0] * IMG_EMB_DIM[1], IMG_EMB_DIM[2])
IMG_SHAPE = (299, 299, 3)
PE_INPUT = IMG_EMB_DIM[0]
IMG_SEQ_LEN, IMG_EMB_DEPTH = IMG_EMB_DIM
D_MODEL = IMG_EMB_DEPTH

# Network parameters
N_LAYERS = 4
D_MODEL = 512
D_FF = 2048
N_HEADS = 8
DROPOUT_RATE = 0.1

# Misc
MAX_LEN = max_length
VOCAB_LEN = len(tokenizer.word_index)
PE_OUTPUT = MAX_LEN
TARGET_V_SIZE = VOCAB_LEN
REPLICA_BATCH_SIZE = 1


# Config Encoder
PREPROCESSING_FN = tf.keras.applications.efficientnet.preprocess_input
BB_FN = Efficient_Net_encoder.get_efficientnetv2_backbone

# Config Model
testing_config = config.Config()

testing_config.initialize_encoder_config(
    image_embedding_dim=IMG_EMB_DIM,
    preprocessing_fn=PREPROCESSING_FN,
    backbone_fn=BB_FN,
    image_shape=IMG_SHAPE,
    do_permute=IMG_EMB_DIM[1] < IMG_EMB_DIM[0],
)

testing_config.initialize_transformer_config(
    vocab_len=VOCAB_LEN,
    max_len=MAX_LEN,
    n_transformer_layers=N_LAYERS,
    transformer_d_dff=D_FF,
    transformer_n_heads=N_HEADS,
    image_embedding_dim=IMG_EMB_DIM,
)

# Prepare model
optimizer, encoder, transformer = config.prepare_models(
    encoder_config=testing_config.encoder_config,
    transformer_config=testing_config.transformer_config,
    replica_batch_size=REPLICA_BATCH_SIZE,
    verbose=0,
)

# Load trained model checkpoint
checkpoint_path = "checkpoints"
ckpt = tf.train.Checkpoint(
    encoder=encoder, transformer=transformer, optimizer=optimizer
)
ckpt_manager = tf.train.CheckpointManager(ckpt, checkpoint_path, max_to_keep=50)
if ckpt_manager.latest_checkpoint:
    ckpt.restore(tf.train.latest_checkpoint(checkpoint_path))
    start_epoch = int(ckpt_manager.latest_checkpoint.split("-")[-1])


class DECIMER_Predictor(tf.Module):
    """This is a class which takes care of inference. It loads the saved checkpoint and the necessary
    tokenizers. The inference begins with the start token (<start>) and ends when the end token(<end>)
    is met. This class can only work with tf.Tensor objects. The strings shoul gets transformed into np.arrays
    before feeding them into this class.
    """

    def __init__(self, encoder, tokenizer, transformer, max_length):
        """Load the tokenizers, the maximum input and output length and the model.

        Args:
            encoder (tf.keras.model):  The encoder model
            tokenizer (tf.keras.tokenizer): Output tokenizer, defines which charater is assigned to what token
            transformer (tf.keras.model):  The transformer model
            max_length (int): Maximum length of a string which can get predicted
        """
        self.encoder = encoder
        self.tokenizer = tokenizer
        self.transformer = transformer
        self.max_length = max_length

    def __call__(self, Decoded_image):
        """This fuction takes in the Decoded image as input and
        makes the predicted list of tokens and return the tokens as tf.Tensor array.
        Before feeding the input array we must define start and the end tokens.

        Args:
            Decoded_image (tf.Tensor[tf.int32]): Input array in tf.Eagertensor format.

        Returns:
            output (tf.Tensor[tf.int64]): predicted output as an array.
        """

        assert isinstance(Decoded_image, tf.Tensor)
        if len(Decoded_image.shape) == 0:
            Decoded_image = Decoded_image[tf.newaxis]

        _image_batch = tf.expand_dims(Decoded_image, 0)
        _image_embedding = encoder(_image_batch, training=False)

        start_token = tf.cast(
            tf.convert_to_tensor([tokenizer.word_index["<start>"]]), tf.int32
        )
        end_token = tf.cast(
            tf.convert_to_tensor([tokenizer.word_index["<end>"]]), tf.int32
        )

        output_array = tf.TensorArray(dtype=tf.int32, size=0, dynamic_size=True)
        output_array = output_array.write(0, start_token)

        for t in tf.range(max_length):
            output = tf.transpose(output_array.stack())
            combined_mask = Transformer_decoder.create_mask(None, output)

            # predictions.shape == (batch_size, seq_len, vocab_size)
            prediction_batch, _ = transformer(
                _image_embedding, output, training=False, look_ahead_mask=combined_mask
            )

            # select the last word from the seq_len dimension
            predictions = prediction_batch[:, -1:, :]  # (batch_size, 1, vocab_size)

            predicted_id = tf.cast(tf.argmax(predictions, axis=-1), tf.int32)

            output_array = output_array.write(t + 1, predicted_id[0])

            if predicted_id == end_token:
                break
        output = tf.transpose(output_array.stack())

        return output


DECIMER = DECIMER_Predictor(encoder, tokenizer, transformer, MAX_LEN)


class ExportDECIMERPredictor(tf.Module):
    """This class wraps the inference class into a module into tf.Module sub-class, with a tf.function on the __call__ method.
    So we could export the model as a tf.saved_model.
    """

    def __init__(self, DECIMER):
        """Import the translator instance."""
        self.DECIMER = DECIMER

    @tf.function
    def __call__(self, Decoded_Image):
        """This fucntion calls the __call__function from the translator class.
        In the tf.function only the output sentence is returned.
        Thanks to the non-strict execution in tf.function any unnecessary values are never computed.

        Args:
            sentence (tf.Tensor[tf.int32]): Input array in tf.Easgertensor format.

        Returns:
            tf.Tensor[tf.int64]: predicted output as an array.
        """

        result = self.DECIMER(Decoded_Image)

        return result


DECIMER_export = ExportDECIMERPredictor(DECIMER)

tf.saved_model.save(
    DECIMER_export,
    export_dir="DECIMER_Packed_model",
    options=tf.saved_model.SaveOptions(experimental_custom_gradients=True),
)



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""EfficientNet V1 and V2 model configs."""
import functools
import re
from efficientnetv2 import hparams

cfg_register = functools.partial(hparams.register, prefix="cfg:")


class BlockDecoder(object):
    """Block Decoder for readability."""

    def _decode_block_string(self, block_string):
        """Gets a block through a string notation of arguments."""
        assert isinstance(block_string, str)
        ops = block_string.split("_")
        options = {}
        for op in ops:
            splits = re.split(r"(\d.*)", op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

        return hparams.Config(
            kernel_size=int(options["k"]),
            num_repeat=int(options["r"]),
            input_filters=int(options["i"]),
            output_filters=int(options["o"]),
            expand_ratio=int(options["e"]),
            se_ratio=float(options["se"]) if "se" in options else None,
            strides=int(options["s"]),
            conv_type=int(options["c"]) if "c" in options else 0,
        )

    def _encode_block_string(self, block):
        """Encodes a block to a string."""
        args = [
            "r%d" % block.num_repeat,
            "k%d" % block.kernel_size,
            "s%d" % block.strides,
            "e%s" % block.expand_ratio,
            "i%d" % block.input_filters,
            "o%d" % block.output_filters,
            "c%d" % block.conv_type,
            "f%d" % block.fused_conv,
        ]
        if block.se_ratio > 0 and block.se_ratio <= 1:
            args.append("se%s" % block.se_ratio)
        return "_".join(args)

    def decode(self, string_list):
        """Decodes a list of string notations to specify blocks inside the network.

        Args:
          string_list: a list of strings, each string is a notation of block.

        Returns:
          A list of namedtuples to represent blocks arguments.
        """
        assert isinstance(string_list, list)
        blocks_args = []
        for block_string in string_list:
            blocks_args.append(self._decode_block_string(block_string))
        return blocks_args

    def encode(self, blocks_args):
        """Encodes a list of Blocks to a list of strings.

        Args:
          blocks_args: A list of namedtuples to represent blocks arguments.
        Returns:
          a list of strings, each string is a notation of block.
        """
        block_strings = []
        for block in blocks_args:
            block_strings.append(self._encode_block_string(block))
        return block_strings


#################### EfficientNet V1 configs ####################
v1_b0_block_str = [
    "r1_k3_s1_e1_i32_o16_se0.25",
    "r2_k3_s2_e6_i16_o24_se0.25",
    "r2_k5_s2_e6_i24_o40_se0.25",
    "r3_k3_s2_e6_i40_o80_se0.25",
    "r3_k5_s1_e6_i80_o112_se0.25",
    "r4_k5_s2_e6_i112_o192_se0.25",
    "r1_k3_s1_e6_i192_o320_se0.25",
]


efficientnetv1_params = {
    # (width_coefficient, depth_coefficient, resolution, dropout_rate)
    "efficientnet-b0": (1.0, 1.0, 224, 0.2),
    "efficientnet-b1": (1.0, 1.1, 240, 0.2),
    "efficientnet-b2": (1.1, 1.2, 260, 0.3),
    "efficientnet-b3": (1.2, 1.4, 300, 0.3),
    "efficientnet-b4": (1.4, 1.8, 380, 0.4),
    "efficientnet-b5": (1.6, 2.2, 456, 0.4),
    "efficientnet-b6": (1.8, 2.6, 528, 0.5),
    "efficientnet-b7": (2.0, 3.1, 600, 0.5),
    "efficientnet-b8": (2.2, 3.6, 672, 0.5),
    "efficientnet-l2": (4.3, 5.3, 800, 0.5),
}


def efficientnetv1_config(model_name="efficientnet-b0"):
    """EfficientNetV1 model config."""
    width_coefficient, depth_coefficient, isize, dropout_rate = efficientnetv1_params[
        model_name
    ]

    cfg = hparams.Config(
        model=dict(
            model_name=model_name,
            blocks_args=BlockDecoder().decode(v1_b0_block_str),
            width_coefficient=width_coefficient,
            depth_coefficient=depth_coefficient,
            dropout_rate=dropout_rate,
        ),
        eval=dict(isize=isize),
        train=dict(isize=0.8),  # 80% of eval size
        data=dict(augname="effnetv1_autoaug"),
    )
    return cfg


#################### EfficientNet V2 configs ####################
v2_base_block = [  # The baseline config for v2 models.
    "r1_k3_s1_e1_i32_o16_c1",
    "r2_k3_s2_e4_i16_o32_c1",
    "r2_k3_s2_e4_i32_o48_c1",
    "r3_k3_s2_e4_i48_o96_se0.25",
    "r5_k3_s1_e6_i96_o112_se0.25",
    "r8_k3_s2_e6_i112_o192_se0.25",
]


v2_s_block = [  # about base * (width1.4, depth1.8)
    "r2_k3_s1_e1_i24_o24_c1",
    "r4_k3_s2_e4_i24_o48_c1",
    "r4_k3_s2_e4_i48_o64_c1",
    "r6_k3_s2_e4_i64_o128_se0.25",
    "r9_k3_s1_e6_i128_o160_se0.25",
    "r15_k3_s2_e6_i160_o256_se0.25",
]


v2_m_block = [  # about base * (width1.6, depth2.2)
    "r3_k3_s1_e1_i24_o24_c1",
    "r5_k3_s2_e4_i24_o48_c1",
    "r5_k3_s2_e4_i48_o80_c1",
    "r7_k3_s2_e4_i80_o160_se0.25",
    "r14_k3_s1_e6_i160_o176_se0.25",
    "r18_k3_s2_e6_i176_o304_se0.25",
    "r5_k3_s1_e6_i304_o512_se0.25",
]


v2_l_block = [  # about base * (width2.0, depth3.1)
    "r4_k3_s1_e1_i32_o32_c1",
    "r7_k3_s2_e4_i32_o64_c1",
    "r7_k3_s2_e4_i64_o96_c1",
    "r10_k3_s2_e4_i96_o192_se0.25",
    "r19_k3_s1_e6_i192_o224_se0.25",
    "r25_k3_s2_e6_i224_o384_se0.25",
    "r7_k3_s1_e6_i384_o640_se0.25",
]

v2_xl_block = [  # only for 21k pretraining.
    "r4_k3_s1_e1_i32_o32_c1",
    "r8_k3_s2_e4_i32_o64_c1",
    "r8_k3_s2_e4_i64_o96_c1",
    "r16_k3_s2_e4_i96_o192_se0.25",
    "r24_k3_s1_e6_i192_o256_se0.25",
    "r32_k3_s2_e6_i256_o512_se0.25",
    "r8_k3_s1_e6_i512_o640_se0.25",
]
efficientnetv2_params = {
    # (block, width, depth, train_size, eval_size, dropout, randaug, mixup, aug)
    "efficientnetv2-s": (  # 83.9% @ 22M
        v2_s_block,
        1.0,
        1.0,
        300,
        384,
        0.2,
        10,
        0,
        "randaug",
    ),
    "efficientnetv2-m": (  # 85.2% @ 54M
        v2_m_block,
        1.0,
        1.0,
        384,
        480,
        0.3,
        15,
        0.2,
        "randaug",
    ),
    "efficientnetv2-l": (  # 85.7% @ 120M
        v2_l_block,
        1.0,
        1.0,
        384,
        480,
        0.4,
        20,
        0.5,
        "randaug",
    ),
    "efficientnetv2-xl": (v2_xl_block, 1.0, 1.0, 384, 512, 0.4, 20, 0.5, "randaug"),
    # For fair comparison to EfficientNetV1, using the same scaling and autoaug.
    "efficientnetv2-b0": (  # 78.7% @ 7M params
        v2_base_block,
        1.0,
        1.0,
        192,
        224,
        0.2,
        0,
        0,
        "effnetv1_autoaug",
    ),
    "efficientnetv2-b1": (  # 79.8% @ 8M params
        v2_base_block,
        1.0,
        1.1,
        192,
        240,
        0.2,
        0,
        0,
        "effnetv1_autoaug",
    ),
    "efficientnetv2-b2": (  # 80.5% @ 10M params
        v2_base_block,
        1.1,
        1.2,
        208,
        260,
        0.3,
        0,
        0,
        "effnetv1_autoaug",
    ),
    "efficientnetv2-b3": (  # 82.1% @ 14M params
        v2_base_block,
        1.2,
        1.4,
        240,
        300,
        0.3,
        0,
        0,
        "effnetv1_autoaug",
    ),
}


def efficientnetv2_config(model_name="efficientnetv2-s"):
    """EfficientNetV2 model config."""
    (
        block,
        width,
        depth,
        train_size,
        eval_size,
        dropout,
        randaug,
        mix,
        aug,
    ) = efficientnetv2_params[model_name]

    cfg = hparams.Config(
        model=dict(
            model_name=model_name,
            blocks_args=BlockDecoder().decode(block),
            width_coefficient=width,
            depth_coefficient=depth,
            dropout_rate=dropout,
        ),
        train=dict(isize=train_size, stages=4, sched=True),
        eval=dict(isize=eval_size),
        data=dict(augname=aug, ram=randaug, mixup_alpha=mix, cutmix_alpha=mix),
    )
    return cfg


################################################################################
def get_model_config(model_name: str):
    """Main entry for model name to config."""
    if model_name.startswith("efficientnet-"):
        return efficientnetv1_config(model_name)
    if model_name.startswith("efficientnetv2-"):
        return efficientnetv2_config(model_name)
    raise ValueError(f"Unknown model_name {model_name}")



# EfficientNet-V2 config
import tensorflow as tf
import DECIMER.efficientnetv2

BATCH_SIZE_DEBUG = 2
MODEL = "efficientnetv2-b3"  # @param


# Define encoder
def get_efficientnetv2_backbone(
    model_name, include_top=False, input_shape=(299, 299, 3), pooling=None, weights=None
):
    """Initiate and get the desired Efficient-Net V2 backbone as encoder

    Args:
        model_name (str): Name of the Efficient-Net V2 model
        include_top (bool, optional):  Defaults to False.
        input_shape (tuple, optional): Image shape. Defaults to (299, 299, 3).
        pooling (int, optional): Max pooling values. Defaults to None.
        weights ( optional): Pretrained weights. Defaults to None.

    Raises:
        NotImplementedError: At this time we only want to use the raw

    Returns:
        Efficient Net V2 backbone
    """
    # Catch unsupported arguments
    if pooling or weights or include_top:
        raise NotImplementedError(
            "\n...At this time we only want to use the raw "
            "(no pretraining), headless, features with no pooling ...\n"
        )
    backbone = efficientnetv2.effnetv2_model.EffNetV2Model(model_name=model_name)
    backbone(
        tf.ones((BATCH_SIZE_DEBUG, *input_shape)), training=False, features_only=True
    )
    return backbone


class Encoder(tf.keras.Model):
    """Encoder class

    Args:
        tf (_type_): tensorflow model module
    """

    def __init__(
        self,
        image_embedding_dim,
        preprocessing_fn,
        backbone_fn,
        image_shape,
        do_permute=False,
        include_top=False,
        pretrained_weights=None,
        scale_factor=0,
    ):

        super(Encoder, self).__init__()

        self.image_embedding_dim = image_embedding_dim
        self.preprocessing_fn = preprocessing_fn
        self.encoder_backbone = backbone_fn(
            model_name=MODEL,
            include_top=include_top,
            weights=pretrained_weights,
            input_shape=image_shape,
        )
        self.reshape = tf.keras.layers.Reshape(
            self.image_embedding_dim, name="image_embedding"
        )
        self.include_top = include_top
        self.scale_factor = scale_factor

    def call(self, x, training):
        x = self.preprocessing_fn(x)
        x = self.encoder_backbone(
            x, training=training, features_only=not self.include_top
        )[self.scale_factor]
        x = self.reshape(x, training=training)
        return x



# -*- coding: utf-8 -*-

"""
DECIMER V2.1.3 Python Package.
============================

This repository contains DECIMER-V2,Deep lEarning for Chemical ImagE Recognition) project
was launched to address the OCSR problem with the latest computational intelligence methods
to provide an automated open-source software solution.


Typical usage example:

from decimer import predict_SMILES

# Chemical depiction to SMILES translation
image_path = "path/to/imagefile"
SMILES = predict_SMILES(image_path)
print(SMILES)

For comments, bug reports or feature ideas,
please raise a issue on the Github repository.

"""

__version__ = "2.1.3"

__all__ = [
    "DECIMER",
]


from .decimer import predict_SMILES



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="decimer",
    version="2.1.3",
    author="Kohulan Rajan",
    author_email="kohulan.rajan@uni-jena.de",
    maintainer="Kohulan Rajan",
    maintainer_email="kohulan.rajan@uni-jena.de",
    description="DECIMER 2.0: Deep Learning for Chemical Image Recognition using Efficient-Net V2 + Transformer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kohulan/DECIMER-Image_Transformer",
    packages=setuptools.find_packages(),
    license="MIT",
    install_requires=[
        "tensorflow>=2.8.0",
        "opencv-python",
        "pystow",
        "pyheif",
        "efficientnet",
    ],
    package_data={"DECIMER": ["repack/*.*", "efficientnetv2/*.*", "Utils/*.*"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
)



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Model utilities."""
import contextlib
import functools
import os
from absl import logging
import numpy as np
import tensorflow as tf
import tensorflow_addons.layers as tfa_layers

from tensorflow.python.tpu import (
    tpu_function,
)  # pylint:disable=g-direct-tensorflow-import


def activation_fn(features: tf.Tensor, act_fn: str):
    """Customized non-linear activation type."""
    if act_fn in ("silu", "swish"):
        return tf.nn.swish(features)
    elif act_fn == "silu_native":
        return features * tf.sigmoid(features)
    elif act_fn == "hswish":
        return features * tf.nn.relu6(features + 3) / 6
    elif act_fn == "relu":
        return tf.nn.relu(features)
    elif act_fn == "relu6":
        return tf.nn.relu6(features)
    elif act_fn == "elu":
        return tf.nn.elu(features)
    elif act_fn == "leaky_relu":
        return tf.nn.leaky_relu(features)
    elif act_fn == "selu":
        return tf.nn.selu(features)
    elif act_fn == "mish":
        return features * tf.math.tanh(tf.math.softplus(features))
    else:
        raise ValueError("Unsupported act_fn {}".format(act_fn))


def get_act_fn(act_fn):
    if not act_fn:
        return tf.nn.silu
    if isinstance(act_fn, str):
        return functools.partial(activation_fn, act_fn=act_fn)
    return act_fn


def cross_replica_mean(t, num_shards_per_group=None):
    """Calculates the average value of input tensor across TPU replicas."""
    num_shards = tpu_function.get_tpu_context().number_of_shards
    if not num_shards_per_group:
        return tf.compat.v1.tpu.cross_replica_sum(t) / tf.cast(num_shards, t.dtype)

    group_assignment = None
    if num_shards_per_group > 1:
        if num_shards % num_shards_per_group != 0:
            raise ValueError(
                "num_shards: %d mod shards_per_group: %d, should be 0"
                % (num_shards, num_shards_per_group)
            )
        num_groups = num_shards // num_shards_per_group
        group_assignment = [
            [x for x in range(num_shards) if x // num_shards_per_group == y]
            for y in range(num_groups)
        ]
    return tf.compat.v1.tpu.cross_replica_sum(t, group_assignment) / tf.cast(
        num_shards_per_group, t.dtype
    )


class WarmupLearningRateSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Provides a variety of learning rate decay schedules with warm up."""

    def __init__(
        self,
        initial_lr,
        steps_per_epoch=None,
        lr_decay_type="exponential",
        decay_factor=0.97,
        decay_epochs=2.4,
        total_steps=None,
        warmup_epochs=5,
        minimal_lr=0,
    ):
        super(WarmupLearningRateSchedule, self).__init__()
        self.initial_lr = initial_lr
        self.steps_per_epoch = steps_per_epoch
        self.lr_decay_type = lr_decay_type
        self.decay_factor = decay_factor
        self.decay_epochs = decay_epochs
        self.total_steps = total_steps
        self.warmup_epochs = warmup_epochs
        self.minimal_lr = minimal_lr

    def __call__(self, step):
        if self.lr_decay_type == "exponential":
            assert self.steps_per_epoch is not None
            decay_steps = self.steps_per_epoch * self.decay_epochs
            lr = tf.keras.optimizers.schedules.ExponentialDecay(
                self.initial_lr, decay_steps, self.decay_factor, staircase=True
            )(step)
        elif self.lr_decay_type == "cosine":
            assert self.total_steps is not None
            lr = (
                0.5
                * self.initial_lr
                * (1 + tf.cos(np.pi * tf.cast(step, tf.float32) / self.total_steps))
            )
        elif self.lr_decay_type == "linear":
            assert self.total_steps is not None
            lr = (1.0 - tf.cast(step, tf.float32) / self.total_steps) * self.initial_lr
        elif self.lr_decay_type == "constant":
            lr = self.initial_lr
        else:
            assert False, "Unknown lr_decay_type : %s" % self.lr_decay_type

        if self.minimal_lr:
            lr = tf.math.maximum(lr, self.minimal_lr)

        if self.warmup_epochs:
            logging.info("Learning rate warmup_epochs: %s", str(self.warmup_epochs))
            warmup_steps = int(self.warmup_epochs * self.steps_per_epoch)
            warmup_lr = (
                self.initial_lr
                * tf.cast(step, tf.float32)
                / tf.cast(warmup_steps, tf.float32)
            )
            lr = tf.cond(step < warmup_steps, lambda: warmup_lr, lambda: lr)

        return lr

    def get_config(self):
        return {
            "initial_lr": self.initial_lr,
            "steps_per_epoch": self.steps_per_epoch,
            "lr_decay_type": self.lr_decay_type,
            "decay_factor": self.decay_factor,
            "decay_epochs": self.decay_epochs,
            "total_steps": self.total_steps,
            "warmup_epochs": self.warmup_epochs,
            "minimal_lr": self.minimal_lr,
        }


def build_optimizer(
    learning_rate, optimizer_name="rmsprop", decay=0.9, epsilon=0.001, momentum=0.9
):
    """Build optimizer."""
    if optimizer_name == "sgd":
        logging.info("Using SGD optimizer")
        optimizer = tf.compat.v1.train.GradientDescentOptimizer(
            learning_rate=learning_rate
        )
    elif optimizer_name == "momentum":
        logging.info("Using Momentum optimizer")
        optimizer = tf.compat.v1.train.MomentumOptimizer(
            learning_rate=learning_rate, momentum=momentum
        )
    elif optimizer_name == "rmsprop":
        logging.info("Using RMSProp optimizer")
        optimizer = tf.compat.v1.train.RMSPropOptimizer(
            learning_rate, decay, momentum, epsilon
        )
    elif optimizer_name == "adam":
        logging.info("Using Adam optimizer")
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate)
    else:
        logging.fatal("Unknown optimizer: %s", optimizer_name)

    return optimizer


class TpuBatchNormalization(tf.keras.layers.BatchNormalization):
    """Cross replica batch normalization."""

    def __init__(self, fused=False, **kwargs):
        if not kwargs.get("name", None):
            kwargs["name"] = "tpu_batch_normalization"
        if fused in (True, None):
            raise ValueError("TpuBatchNormalization does not support fused=True.")
        super().__init__(fused=fused, **kwargs)

    def _moments(self, inputs, reduction_axes, keep_dims):
        """Compute the mean and variance: it overrides the original _moments."""
        shard_mean, shard_variance = super()._moments(
            inputs, reduction_axes, keep_dims=keep_dims
        )

        num_shards = tpu_function.get_tpu_context().number_of_shards or 1
        num_shards_per_group = min(8, num_shards)  # aggregate up to 32 cores.
        if num_shards_per_group > 1:
            logging.info(
                "TpuBatchNormalization with num_shards_per_group %d",
                num_shards_per_group,
            )
            # Compute variance using: Var[X]= E[X^2] - E[X]^2.
            shard_square_of_mean = tf.math.square(shard_mean)
            shard_mean_of_square = shard_variance + shard_square_of_mean
            group_mean = cross_replica_mean(shard_mean, num_shards_per_group)
            group_mean_of_square = cross_replica_mean(
                shard_mean_of_square, num_shards_per_group
            )
            group_variance = group_mean_of_square - tf.math.square(group_mean)
            return (group_mean, group_variance)
        else:
            return (shard_mean, shard_variance)

    def call(self, inputs, training=None):
        outputs = super().call(inputs, training)
        # A temporary hack for tf1 compatibility with keras batch norm.
        for u in self.updates:
            tf.compat.v1.add_to_collection(tf.compat.v1.GraphKeys.UPDATE_OPS, u)
        return outputs


class BatchNormalization(tf.keras.layers.BatchNormalization):
    """Fixed default name of BatchNormalization to match TpuBatchNormalization."""

    def __init__(self, **kwargs):
        if not kwargs.get("name", None):
            kwargs["name"] = "tpu_batch_normalization"
        super().__init__(**kwargs)

    def call(self, inputs, training=None):
        outputs = super().call(inputs, training)
        if training and not tf.executing_eagerly():
            # A temporary hack for tf1 compatibility with keras batch norm.
            for u in self.updates:
                tf.compat.v1.add_to_collection(tf.compat.v1.GraphKeys.UPDATE_OPS, u)
        return outputs


def normalization(
    norm_type: str, axis=-1, epsilon=0.001, momentum=0.99, groups=8, name=None
):
    """Normalization after conv layers."""
    if norm_type == "gn":
        return tfa_layers.GroupNormalization(groups, axis, epsilon, name=name)

    if norm_type == "tpu_bn":
        return TpuBatchNormalization(
            axis=axis, momentum=momentum, epsilon=epsilon, name=name
        )

    return BatchNormalization(axis=axis, momentum=momentum, epsilon=epsilon, name=name)


def archive_ckpt(ckpt_eval, ckpt_objective, ckpt_path):
    """Archive a checkpoint if the metric is better."""
    ckpt_dir, ckpt_name = os.path.split(ckpt_path)

    saved_objective_path = os.path.join(ckpt_dir, "best_objective.txt")
    saved_objective = float("-inf")
    if tf.io.gfile.exists(saved_objective_path):
        with tf.io.gfile.GFile(saved_objective_path, "r") as f:
            saved_objective = float(f.read())
    if saved_objective > ckpt_objective:
        logging.info("Ckpt %s is worse than %s", ckpt_objective, saved_objective)
        return False

    filenames = tf.io.gfile.glob(ckpt_path + ".*")
    if filenames is None:
        logging.info("No files to copy for checkpoint %s", ckpt_path)
        return False

    # Clear the old folder.
    dst_dir = os.path.join(ckpt_dir, "archive")
    if tf.io.gfile.exists(dst_dir):
        tf.io.gfile.rmtree(dst_dir)
    tf.io.gfile.makedirs(dst_dir)

    # Write checkpoints.
    for f in filenames:
        dest = os.path.join(dst_dir, os.path.basename(f))
        tf.io.gfile.copy(f, dest, overwrite=True)
    ckpt_state = tf.compat.v1.train.generate_checkpoint_state_proto(
        dst_dir, model_checkpoint_path=ckpt_name, all_model_checkpoint_paths=[ckpt_name]
    )
    with tf.io.gfile.GFile(os.path.join(dst_dir, "checkpoint"), "w") as f:
        f.write(str(ckpt_state))
    with tf.io.gfile.GFile(os.path.join(dst_dir, "best_eval.txt"), "w") as f:
        f.write("%s" % ckpt_eval)

    # Update the best objective.
    with tf.io.gfile.GFile(saved_objective_path, "w") as f:
        f.write("%f" % ckpt_objective)

    logging.info("Copying checkpoint %s to %s", ckpt_path, dst_dir)
    return True


def get_ema_vars():
    """Get all exponential moving average (ema) variables."""
    ema_vars = tf.compat.v1.trainable_variables() + tf.compat.v1.get_collection(
        "moving_vars"
    )
    for v in tf.compat.v1.global_variables():
        # We maintain mva for batch norm moving mean and variance as well.
        if "moving_mean" in v.name or "moving_variance" in v.name:
            ema_vars.append(v)
    return list(set(ema_vars))


def drop_connect(inputs, is_training, survival_prob):
    """Drop the entire conv with given survival probability."""
    # "Deep Networks with Stochastic Depth", https://arxiv.org/pdf/1603.09382.pdf
    if not is_training:
        return inputs

    # Compute tensor.
    batch_size = tf.shape(inputs)[0]
    random_tensor = survival_prob
    random_tensor += tf.random.uniform([batch_size, 1, 1, 1], dtype=inputs.dtype)
    binary_tensor = tf.floor(random_tensor)
    # Unlike conventional way that multiply survival_prob at test time, here we
    # divide survival_prob at training time, such that no addition compute is
    # needed at test time.
    output = inputs / survival_prob * binary_tensor
    return output


def num_params_flops(readable_format=True):
    """Return number of parameters and flops."""
    nparams = np.sum(
        [np.prod(v.get_shape().as_list()) for v in tf.compat.v1.trainable_variables()]
    )
    options = tf.compat.v1.profiler.ProfileOptionBuilder.float_operation()
    options["output"] = "none"
    flops = tf.compat.v1.profiler.profile(
        tf.compat.v1.get_default_graph(), options=options
    ).total_float_ops
    # We use flops to denote multiply-adds, which is counted as 2 ops in tfprof.
    flops = flops // 2
    if readable_format:
        nparams = float(nparams) * 1e-6
        flops = float(flops) * 1e-9
    return nparams, flops


class Pair(tuple):
    def __new__(cls, name, value):
        return super().__new__(cls, (name, value))

    def __init__(self, name, _):  # pylint: disable=super-init-not-called
        self.name = name


def scalar(name, tensor, is_tpu=True):
    """Stores a (name, Tensor) tuple in a custom collection."""
    logging.info("Adding scale summary %s", Pair(name, tensor))
    if is_tpu:
        tf.compat.v1.add_to_collection(
            "scalar_summaries", Pair(name, tf.reduce_mean(tensor))
        )
    else:
        tf.summary.scalar(name, tf.reduce_mean(tensor))


def image(name, tensor, is_tpu=True):
    logging.info("Adding image summary %s", Pair(name, tensor))
    if is_tpu:
        tf.compat.v1.add_to_collection("image_summaries", Pair(name, tensor))
    else:
        tf.summary.image(name, tensor)


def get_tpu_host_call(global_step, model_dir, iterations_per_loop):
    """Get TPU host call for summaries."""
    scalar_summaries = tf.compat.v1.get_collection("scalar_summaries")
    if not scalar_summaries:
        return None  # No summaries to write.

    def host_call_fn(global_step, *args):
        """Training host call. Creates summaries for training metrics."""
        gs = global_step[0]
        with tf.summary.create_file_writer(
            model_dir, max_queue=iterations_per_loop
        ).as_default():
            with tf.summary.record_if(True):
                for i, _ in enumerate(scalar_summaries):
                    name = scalar_summaries[i][0]
                    tensor = args[i][0]
                    tf.summary.scalar(name, tensor, step=gs)
            return tf.compat.v1.summary.all_v2_summary_ops()

    reshaped_tensors = [tf.reshape(t, [1]) for _, t in scalar_summaries]
    global_step_t = tf.reshape(global_step, [1])
    return host_call_fn, [global_step_t] + reshaped_tensors


@contextlib.contextmanager
def float16_scope():
    """Scope class for float16."""

    def _custom_getter(getter, *args, **kwargs):
        """Returns a custom getter that methods must be called under."""
        cast_to_float16 = False
        requested_dtype = kwargs["dtype"]
        if requested_dtype == tf.float16:
            kwargs["dtype"] = tf.float32
            cast_to_float16 = True
        var = getter(*args, **kwargs)
        if cast_to_float16:
            var = tf.cast(var, tf.float16)
        return var

    with tf.compat.v1.variable_scope("", custom_getter=_custom_getter) as varscope:
        yield varscope


def set_precision_policy(policy_name=None, loss_scale=False):
    """Set precision policy according to the name.

    Args:
      policy_name: precision policy name, one of 'float32', 'mixed_float16',
        'mixed_bfloat16', or None.
      loss_scale: whether to use loss scale (only for training).
    """
    if not policy_name:
        return

    assert policy_name in ("mixed_float16", "mixed_bfloat16", "float32")
    logging.info("use mixed precision policy name %s", policy_name)
    tf.compat.v1.keras.layers.enable_v2_dtype_behavior()
    # mixed_float16 training is not supported for now, so disable loss_scale.
    # float32 and mixed_bfloat16 do not need loss scale for training.
    if loss_scale:
        policy = tf.keras.mixed_precision.experimental.Policy(policy_name)
    else:
        policy = tf.keras.mixed_precision.experimental.Policy(
            policy_name, loss_scale=None
        )
    tf.keras.mixed_precision.experimental.set_policy(policy)


def build_model_with_precision(pp, mm, ii, tt, *args, **kwargs):
    """Build model with its inputs/params for a specified precision context.

    This is highly specific to this codebase, and not intended to be general API.
    Advanced users only. DO NOT use it if you don't know what it does.
    NOTE: short argument names are intended to avoid conficts with kwargs.

    Args:
      pp: A string, precision policy name, such as "mixed_float16".
      mm: A function, for rmodel builder.
      ii: A tensor, for model inputs.
      tt: A bool, If true, it is for training; otherwise, it is for eval.
      *args: A list of model arguments.
      **kwargs: A dict, extra model parameters.

    Returns:
      the output of mm model.
    """
    if pp == "mixed_bfloat16":
        set_precision_policy(pp)
        inputs = tf.cast(ii, tf.bfloat16)
        with tf.compat.v1.tpu.bfloat16_scope():
            outputs = mm(inputs, *args, **kwargs)
        set_precision_policy("float32")
    elif pp == "mixed_float16":
        set_precision_policy(pp, loss_scale=tt)
        inputs = tf.cast(ii, tf.float16)
        with float16_scope():
            outputs = mm(inputs, *args, **kwargs)
        set_precision_policy("float32")
    elif not pp or pp == "float32":
        outputs = mm(ii, *args, **kwargs)
    else:
        raise ValueError("Unknow precision name {}".format(pp))

    # Users are responsible to convert the dtype of all outputs.
    return outputs


def get_ckpt_var_map(
    ckpt_path, ckpt_scope="", var_scope="", skip_mismatch=None, init_ema=True
):
    """Get a var map for restoring from pretrained checkpoints.

    Args:
      ckpt_path: string. A pretrained checkpoint path.
      ckpt_scope: string. Scope name for checkpoint variables.
      var_scope: string. Scope name for model variables.
      skip_mismatch: skip variables if shape mismatch.
      init_ema: If true, try to init from ema variables.

    Returns:
      var_map: a dictionary from checkpoint name to model variables.
    """
    logging.info("Init model from checkpoint %s", ckpt_path)
    var_map = {}
    # Get the list of vars to restore.
    model_vars = tf.compat.v1.get_collection(
        tf.compat.v1.GraphKeys.GLOBAL_VARIABLES, scope=var_scope
    )
    reader = tf.train.load_checkpoint(ckpt_path)
    ckpt_var_name_to_shape = reader.get_variable_to_shape_map()
    ckpt_var_names = set(reader.get_variable_to_shape_map().keys())
    for v in model_vars:
        v_name = v.op.name

        # filter special variables.
        flist = ["global_step", "ExponentialMovingAverage", "Momentum", "RMSProp"]
        if list(filter(lambda x, s=v_name: x in s, flist)):
            continue

        if not v.op.name.startswith(var_scope):
            logging.info("skip %s -- does not match scope %s", v_name, var_scope)
        cv_name = ckpt_scope + v.op.name[len(var_scope) :]
        if init_ema and cv_name + "/ExponentialMovingAverage" in ckpt_var_names:
            cv_name = cv_name + "/ExponentialMovingAverage"  # prefer ema vars.

        if cv_name not in ckpt_var_names:
            if skip_mismatch:
                logging.info("skip %s (%s) -- not in ckpt", v_name, cv_name)
                continue
            raise ValueError(f"{v.op} is not in ckpt {ckpt_path}")

        cv_shape = ckpt_var_name_to_shape[cv_name]
        if v.shape != cv_shape:
            if skip_mismatch:
                logging.info(
                    "skip %s (%s vs %s) -- shape mismatch", v_name, v.shape, cv_shape
                )
                continue
            raise ValueError(f"shape mismatch {v_name} ({v.shape} vs {cv_shape})")

        var_map[cv_name] = v

    if not var_map or len(var_map) < 5:
        raise ValueError(f"var_map={var_map} is almost empty, please check logs.")

    for (k, v) in var_map.items():
        logging.log_first_n(logging.INFO, f"Init {v.op.name} from ckpt var {k}", 10)

    return var_map



# Network configuration file
import tensorflow as tf
import efficientnet.tfkeras as efn
import DECIMER.Efficient_Net_encoder
import DECIMER.Transformer_decoder
from PIL import Image
import numpy as np
import io
import cv2
import pystow
import pathlib
import pyheif
import zipfile

TARGET_DTYPE = tf.float32


def central_square_image(im):
    """
    This function takes a Pillow Image object and will add white padding
    so that the image has a square shape with the width/height of the longest side
    of the original image.
    ___
    im: PIL.Image
    ___
    output: PIL.Image
    """
    max_wh = int(1.2 * max(im.size))
    # If the new image is smaller than 299x299, then let's paste it into an empty image
    # of that size instead of distorting it later while resizing.
    if max_wh < 299:
        max_wh = 299
    new_im = Image.new(im.mode, (max_wh, max_wh), "white")
    paste_pos = (
        int((new_im.size[0] - im.size[0]) / 2),
        int((new_im.size[1] - im.size[1]) / 2),
    )
    new_im.paste(im, paste_pos)
    return new_im


def delete_empty_borders(im):
    """This function takes a Pillow Image object, converts it to grayscale and
    deletes white space at the borders.
    ___
    im: PIL.Image
    ___
    output: PIL.Image
    """
    im = np.asarray(im.convert("L"))
    mask = im > 200
    rows = np.flatnonzero((~mask).sum(axis=1))
    cols = np.flatnonzero((~mask).sum(axis=0))
    crop = im[rows.min() : rows.max() + 1, cols.min() : cols.max() + 1]
    return Image.fromarray(crop)


def PIL_im_to_BytesIO(im):
    """
    Convert pillow image to io.BytesIO object
    ___
    im: PIL.Image
    ___
    Output: io.BytesIO object with the image data
    """
    output = io.BytesIO()
    im.save(output, format="PNG")
    return output


def HEIF_to_pillow(image_path: str):
    """
    Converts Apple's HEIF format to useful pillow object
    ___
    image_path (str): path of input image
    ___
    Output: PIL.Image
    """
    heif_file = pyheif.read(image_path)
    pil_im = Image.frombytes(
        heif_file.mode,
        heif_file.size,
        heif_file.data,
        "raw",
        heif_file.mode,
        heif_file.stride,
    )
    return pil_im.convert("RGBA")


def remove_transparent(image_path: str):
    """
    Removes the transparent layer from a PNG image with an alpha channel
    ___
    image_path (str): path of input image
    ___
    Output: PIL.Image
    """
    if pathlib.Path(image_path).suffix == ".HEIC":
        png = HEIF_to_pillow(image_path)
    else:
        png = Image.open(image_path).convert("RGBA")

    background = Image.new("RGBA", png.size, (255, 255, 255))
    alpha_composite = Image.alpha_composite(background, png)

    return alpha_composite


def get_bnw_image(image):
    """
    converts images to black and white
    ___
    image: PIL.Image
    ___
    Output: PIL.Image
    """

    im_np = np.asarray(image)
    grayscale = cv2.cvtColor(im_np, cv2.COLOR_BGR2GRAY)
    (thresh, im_bw) = cv2.threshold(
        grayscale, 128, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU
    )
    im_pil = Image.fromarray(im_bw)
    return im_pil


def decode_image(image_path: str):
    """
    Loads and preprocesses an image
    Args:
        image_path (str): path of input image

    Returns:
        Processed image
    """
    img = remove_transparent(image_path)
    img = get_bnw_image(img)
    img = delete_empty_borders(img)
    img = central_square_image(img)
    img = PIL_im_to_BytesIO(img)
    img = tf.image.decode_png(img.getvalue(), channels=3)
    img = tf.image.resize(img, (299, 299))
    img = efn.preprocess_input(img)
    return img


class Config:
    """
    Configuration class
    """

    def __init__(
        self,
    ):
        self.encoder_config = {}
        self.transformer_config = {}
        self.lr_config = {}

    def initialize_encoder_config(
        self,
        image_embedding_dim,
        preprocessing_fn,
        backbone_fn,
        image_shape,
        do_permute=False,
        pretrained_weights=None,
    ):
        """This functions initializes the Efficient-Net V2 encoder with user defined
        configurations.

        Args:
            image_embedding_dim (int): Embedding dimention of the input image
            preprocessing_fn (method): Efficient Net preprocessing function for input image
            backbone_fn (method): Calls Efficient-Net V2 as backbone for encoder
            image_shape (int): Shape of the input image
            do_permute (bool, optional): . Defaults to False.
            pretrained_weights (keras weights, optional): Use pretrainined efficient net weights or not. Defaults to None.
        """
        self.encoder_config = dict(
            image_embedding_dim=image_embedding_dim,
            preprocessing_fn=preprocessing_fn,
            backbone_fn=backbone_fn,
            image_shape=image_shape,
            do_permute=do_permute,
            pretrained_weights=pretrained_weights,
        )

    def initialize_transformer_config(
        self,
        vocab_len,
        max_len,
        n_transformer_layers,
        transformer_d_dff,
        transformer_n_heads,
        image_embedding_dim,
        dropout_rate=0.1,
    ):
        """This functions initializes the Transformer model as decoder with user defined
        configurations.


        Args:
            vocab_len (int): Total number of words in the input vocabulary
            max_len (int): Maximum length of the string found on the training dataset
            n_transformer_layers (int): Number of layers present in the transformer model
            transformer_d_dff (int): Transformer feed forward upwards projection size
            transformer_n_heads (int): Number of heads present in the transformer model
            image_embedding_dim (int): Total number of dimension the image gets embeddeded
            dropout_rate (float, optional): Fraction of the input units to drop. Defaults to 0.1.
        """
        self.transformer_config = dict(
            num_layers=n_transformer_layers,
            d_model=image_embedding_dim[-1],
            num_heads=transformer_n_heads,
            dff=transformer_d_dff,
            target_vocab_size=vocab_len,
            pe_input=image_embedding_dim[0],
            pe_target=max_len,
            dropout_rate=0.1,
        )

    def initialize_lr_config(self, warm_steps, n_epochs):
        """This function sets the configuration to initialize learning rate

        Args:
            warm_steps (int): Number of steps The learning rate is increased
            n_epochs (int): Number of epochs
        """
        self.lr_config = dict(
            warm_steps=warm_steps,
            n_epochs=n_epochs,
        )


class CustomSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Custom schedule for learning rate used during training

    Args:
        tf (_type_): keras learning rate schedule
    """

    def __init__(self, d_model, warmup_steps=4000):
        super(CustomSchedule, self).__init__()

        self.d_model = d_model
        self.d_model = tf.cast(self.d_model, tf.float32)

        self.warmup_steps = warmup_steps

    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps**-1.5)

        return tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)


def prepare_models(encoder_config, transformer_config, replica_batch_size, verbose=0):
    """This function is used to initiate the Encoder and the Transformer with appropriate
    configs set by the user. After initiating the models this function returns the Encoder,Transformer
    and the optimizer.

    Args:
        encoder_config ([type]): Encoder configuration set by user in the config class.
        transformer_config ([type]): Transformer configuration set by user in the config class.
        replica_batch_size ([type]): Per replica batch size set by user(during distributed training).
        verbose (int, optional): Defaults to 0.

    Returns:
        [type]: Optimizer, Encoder model and the Transformer
    """

    # Instiate an optimizer
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.00051)

    # Instantiate the encoder model

    encoder = Efficient_Net_encoder.Encoder(**encoder_config)
    initialization_batch = encoder(
        tf.ones(
            ((replica_batch_size,) + encoder_config["image_shape"]), dtype=TARGET_DTYPE
        ),
        training=False,
    )

    # Instantiate the decoder model
    transformer = Transformer_decoder.Transformer(**transformer_config)
    transformer(
        initialization_batch, tf.random.uniform((replica_batch_size, 1)), training=False
    )

    # Show the model architectures and plot the learning rate
    if verbose:
        print("\nEncoder model\n")
        print(encoder.summary())

        print("\nTransformer model\n")
        print(transformer.summary())

    return optimizer, encoder, transformer


# Downloads the model and unzips the file downloaded, if the model is not present on the working directory.
def download_trained_weights(model_url: str, model_path: str, verbose=1):
    """This function downloads the trained models and tokenizers to a default location.
    After downloading the zipped file the function unzips the file automatically.
    If the model exists on the default location this function will not work.
    Args:
        model_url (str): trained model url for downloading.
        model_path (str): model default path to download.
    Returns:
        downloaded model.
    """
    # Download trained models
    if verbose > 0:
        print("Downloading trained model to " + str(model_path))
        model_path = pystow.ensure("DECIMER-V2", url=model_url)
        print(model_path)
    if verbose > 0:
        print("... done downloading trained model!")
        with zipfile.ZipFile(model_path.as_posix(), "r") as zip_ref:
            zip_ref.extractall(model_path.parent.as_posix())



import os
import sys
import tensorflow as tf

import pickle
from selfies import decoder
import Transformer_decoder
import Efficient_Net_encoder
import config

# Set GPU
os.environ["CUDA_VISIBLE_DEVICES"] = "2"
gpus = tf.config.experimental.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)

# load assets
HERE = os.path.dirname(os.path.abspath(__file__))
tokenizer = pickle.load(
    open(os.path.join(HERE, "tokenizer_Isomeric_SELFIES.pkl"), "rb")
)
max_length = pickle.load(
    open(os.path.join(HERE, "max_length_Isomeric_SELFIES.pkl"), "rb")
)

# Image partameters
IMG_EMB_DIM = (10, 10, 232)
IMG_EMB_DIM = (IMG_EMB_DIM[0] * IMG_EMB_DIM[1], IMG_EMB_DIM[2])
IMG_SHAPE = (299, 299, 3)
PE_INPUT = IMG_EMB_DIM[0]
IMG_SEQ_LEN, IMG_EMB_DEPTH = IMG_EMB_DIM
D_MODEL = IMG_EMB_DEPTH

# Network parameters
N_LAYERS = 4
D_MODEL = 512
D_FF = 2048
N_HEADS = 8
DROPOUT_RATE = 0.1

# Misc
MAX_LEN = max_length
VOCAB_LEN = len(tokenizer.word_index)
PE_OUTPUT = MAX_LEN
TARGET_V_SIZE = VOCAB_LEN
REPLICA_BATCH_SIZE = 1

# Config Encoder
PREPROCESSING_FN = tf.keras.applications.efficientnet.preprocess_input
BB_FN = Efficient_Net_encoder.get_efficientnetv2_backbone

# Config Model
testing_config = config.Config()

testing_config.initialize_encoder_config(
    image_embedding_dim=IMG_EMB_DIM,
    preprocessing_fn=PREPROCESSING_FN,
    backbone_fn=BB_FN,
    image_shape=IMG_SHAPE,
    do_permute=IMG_EMB_DIM[1] < IMG_EMB_DIM[0],
)

testing_config.initialize_transformer_config(
    vocab_len=VOCAB_LEN,
    max_len=MAX_LEN,
    n_transformer_layers=N_LAYERS,
    transformer_d_dff=D_FF,
    transformer_n_heads=N_HEADS,
    image_embedding_dim=IMG_EMB_DIM,
)

# print(f"Encoder config:\n\t -> {testing_config.encoder_config}\n")
# print(f"Transformer config:\n\t -> {testing_config.transformer_config}\n")

# Prepare model
optimizer, encoder, transformer = config.prepare_models(
    encoder_config=testing_config.encoder_config,
    transformer_config=testing_config.transformer_config,
    replica_batch_size=REPLICA_BATCH_SIZE,
    verbose=0,
)

# Load trained model checkpoint
checkpoint_path = os.path.join(HERE, "checkpoints_")
ckpt = tf.train.Checkpoint(
    encoder=encoder, transformer=transformer, optimizer=optimizer
)
ckpt_manager = tf.train.CheckpointManager(ckpt, checkpoint_path, max_to_keep=50)

start_epoch = 0
if ckpt_manager.latest_checkpoint:
    ckpt.restore(tf.train.latest_checkpoint(checkpoint_path))
    start_epoch = int(ckpt_manager.latest_checkpoint.split("-")[-1])


def main():
    if len(sys.argv) != 2:
        print("Usage: {} $image_path".format(sys.argv[0]))
    else:
        SMILES = predict_SMILES(sys.argv[1])
        print(SMILES)


def evaluate(image_path: str):
    """
    This function takes an image path (str) and returns the SELFIES
    representation of the depicted molecule (str).

    Args:
        image_path (str): Path of chemical structure depiction image

    Returns:
        (str): SELFIES representation of the molecule in the input image
    """
    sample = config.decode_image(image_path)
    _image_batch = tf.expand_dims(sample, 0)
    _image_embedding = encoder(_image_batch, training=False)
    output = tf.expand_dims([tokenizer.word_index["<start>"]], 0)
    result = []
    end_token = tokenizer.word_index["<end>"]

    for i in range(MAX_LEN):
        combined_mask = Transformer_decoder.create_mask(None, output)
        prediction_batch, _ = transformer(
            _image_embedding, output, training=False, look_ahead_mask=combined_mask
        )

        predictions = prediction_batch[:, -1:, :]
        predicted_id = tf.cast(tf.argmax(predictions, axis=-1), tf.int32)

        if predicted_id == end_token:
            return result

        result.append(tokenizer.index_word[int(predicted_id)])
        output = tf.concat([output, predicted_id], axis=-1)

    return result


def predict_SMILES(image_path: str):
    """
    This function takes an image path (str) and returns the SMILES
    representation of the depicted molecule (str).

    Args:
        image_path (str): Path of chemical structure depiction image

    Returns:
        (str): SMILES representation of the molecule in the input image
    """
    predicted_SELFIES = evaluate(image_path)
    predicted_SMILES = decoder(
        "".join(predicted_SELFIES).replace("<start>", "").replace("<end>", "")
    )

    return predicted_SMILES


if __name__ == "__main__":
    main()
